# encoding: utf-8

from enum import Enum
from bisect import bisect

from functools import lru_cache as cache

""" from https://unicode-table.com/en/blocks/
"""

# Define type aliases
Char = str


unicode_mapping = [
(0x0000,0x007F,"Basic Latin"),
(0x0080,0x00FF,"Latin-1 Supplement"),
(0x0100,0x017F,"Latin Extended-A"),
(0x0180,0x024F,"Latin Extended-B"),
(0x0250,0x02AF,"IPA Extensions"),
(0x02B0,0x02FF,"Spacing Modifier Letters"),
(0x0300,0x036F,"Combining Diacritical Marks"),
(0x0370,0x03FF,"Greek and Coptic"),
(0x0400,0x04FF,"Cyrillic"),
(0x0500,0x052F,"Cyrillic Supplement"),
(0x0530,0x058F,"Armenian"),
(0x0590,0x05FF,"Hebrew"),
(0x0600,0x06FF,"Arabic"),
(0x0700,0x074F,"Syriac"),
(0x0750,0x077F,"Arabic Supplement"),
(0x0780,0x07BF,"Thaana"),
(0x07C0,0x07FF,"NKo"),
(0x0800,0x083F,"Samaritan"),
(0x0840,0x085F,"Mandaic"),
(0x0860,0x086F,"Syriac Supplement"),
(0x08A0,0x08FF,"Arabic Extended-A"),
(0x0900,0x097F,"Devanagari"),
(0x0980,0x09FF,"Bengali"),
(0x0A00,0x0A7F,"Gurmukhi"),
(0x0A80,0x0AFF,"Gujarati"),
(0x0B00,0x0B7F,"Oriya"),
(0x0B80,0x0BFF,"Tamil"),
(0x0C00,0x0C7F,"Telugu"),
(0x0C80,0x0CFF,"Kannada"),
(0x0D00,0x0D7F,"Malayalam"),
(0x0D80,0x0DFF,"Sinhala"),
(0x0E00,0x0E7F,"Thai"),
(0x0E80,0x0EFF,"Lao"),
(0x0F00,0x0FFF,"Tibetan"),
(0x1000,0x109F,"Myanmar"),
(0x10A0,0x10FF,"Georgian"),
(0x1100,0x11FF,"Hangul Jamo"),
(0x1200,0x137F,"Ethiopic"),
(0x1380,0x139F,"Ethiopic Supplement"),
(0x13A0,0x13FF,"Cherokee"),
(0x1400,0x167F,"Unified Canadian Aboriginal Syllabics"),
(0x1680,0x169F,"Ogham"),
(0x16A0,0x16FF,"Runic"),
(0x1700,0x171F,"Tagalog"),
(0x1720,0x173F,"Hanunoo"),
(0x1740,0x175F,"Buhid"),
(0x1760,0x177F,"Tagbanwa"),
(0x1780,0x17FF,"Khmer"),
(0x1800,0x18AF,"Mongolian"),
(0x18B0,0x18FF,"Unified Canadian Aboriginal Syllabics Extended"),
(0x1900,0x194F,"Limbu"),
(0x1950,0x197F,"Tai Le"),
(0x1980,0x19DF,"New Tai Lue"),
(0x19E0,0x19FF,"Khmer Symbols"),
(0x1A00,0x1A1F,"Buginese"),
(0x1A20,0x1AAF,"Tai Tham"),
(0x1AB0,0x1AFF,"Combining Diacritical Marks Extended"),
(0x1B00,0x1B7F,"Balinese"),
(0x1B80,0x1BBF,"Sundanese"),
(0x1BC0,0x1BFF,"Batak"),
(0x1C00,0x1C4F,"Lepcha"),
(0x1C50,0x1C7F,"Ol Chiki"),
(0x1C80,0x1C8F,"Cyrillic Extended C"),
(0x1C90,0x1CBF,"Georgian Extended"),
(0x1CC0,0x1CCF,"Sundanese Supplement"),
(0x1CD0,0x1CFF,"Vedic Extensions"),
(0x1D00,0x1D7F,"Phonetic Extensions"),
(0x1D80,0x1DBF,"Phonetic Extensions Supplement"),
(0x1DC0,0x1DFF,"Combining Diacritical Marks Supplement"),
(0x1E00,0x1EFF,"Latin Extended Additional"),
(0x1F00,0x1FFF,"Greek Extended"),
(0x2000,0x206F,"General Punctuation"),
(0x2070,0x209F,"Superscripts and Subscripts"),
(0x20A0,0x20CF,"Currency Symbols"),
(0x20D0,0x20FF,"Combining Diacritical Marks for Symbols"),
(0x2100,0x214F,"Letterlike Symbols"),
(0x2150,0x218F,"Number Forms"),
(0x2190,0x21FF,"Arrows"),
(0x2200,0x22FF,"Mathematical Operators"),
(0x2300,0x23FF,"Miscellaneous Technical"),
(0x2400,0x243F,"Control Pictures"),
(0x2440,0x245F,"Optical Character Recognition"),
(0x2460,0x24FF,"Enclosed Alphanumerics"),
(0x2500,0x257F,"Box Drawing"),
(0x2580,0x259F,"Block Elements"),
(0x25A0,0x25FF,"Geometric Shapes"),
(0x2600,0x26FF,"Miscellaneous Symbols"),
(0x2700,0x27BF,"Dingbats"),
(0x27C0,0x27EF,"Miscellaneous Mathematical Symbols-A"),
(0x27F0,0x27FF,"Supplemental Arrows-A"),
(0x2800,0x28FF,"Braille Patterns"),
(0x2900,0x297F,"Supplemental Arrows-B"),
(0x2980,0x29FF,"Miscellaneous Mathematical Symbols-B"),
(0x2A00,0x2AFF,"Supplemental Mathematical Operators"),
(0x2B00,0x2BFF,"Miscellaneous Symbols and Arrows"),
(0x2C00,0x2C5F,"Glagolitic"),
(0x2C60,0x2C7F,"Latin Extended-C"),
(0x2C80,0x2CFF,"Coptic"),
(0x2D00,0x2D2F,"Georgian Supplement"),
(0x2D30,0x2D7F,"Tifinagh"),
(0x2D80,0x2DDF,"Ethiopic Extended"),
(0x2DE0,0x2DFF,"Cyrillic Extended-A"),
(0x2E00,0x2E7F,"Supplemental Punctuation"),
(0x2E80,0x2EFF,"CJK Radicals Supplement"),
(0x2F00,0x2FDF,"Kangxi Radicals"),
(0x2FF0,0x2FFF,"Ideographic Description Characters"),
(0x3000,0x303F,"CJK Symbols and Punctuation"),
(0x3040,0x309F,"Hiragana"),
(0x30A0,0x30FF,"Katakana"),
(0x3100,0x312F,"Bopomofo"),
(0x3130,0x318F,"Hangul Compatibility Jamo"),
(0x3190,0x319F,"Kanbun"),
(0x31A0,0x31BF,"Bopomofo Extended"),
(0x31C0,0x31EF,"CJK Strokes"),
(0x31F0,0x31FF,"Katakana Phonetic Extensions"),
(0x3200,0x32FF,"Enclosed CJK Letters and Months"),
(0x3300,0x33FF,"CJK Compatibility"),
(0x3400,0x4DBF,"CJK Unified Ideographs Extension A"),
(0x4DC0,0x4DFF,"Yijing Hexagram Symbols"),
(0x4E00,0x9FFF,"CJK Unified Ideographs"),
(0xA000,0xA48F,"Yi Syllables"),
(0xA490,0xA4CF,"Yi Radicals"),
(0xA4D0,0xA4FF,"Lisu"),
(0xA500,0xA63F,"Vai"),
(0xA640,0xA69F,"Cyrillic Extended-B"),
(0xA6A0,0xA6FF,"Bamum"),
(0xA700,0xA71F,"Modifier Tone Letters"),
(0xA720,0xA7FF,"Latin Extended-D"),
(0xA800,0xA82F,"Syloti Nagri"),
(0xA830,0xA83F,"Common Indic Number Forms"),
(0xA840,0xA87F,"Phags-pa"),
(0xA880,0xA8DF,"Saurashtra"),
(0xA8E0,0xA8FF,"Devanagari Extended"),
(0xA900,0xA92F,"Kayah Li"),
(0xA930,0xA95F,"Rejang"),
(0xA960,0xA97F,"Hangul Jamo Extended-A"),
(0xA980,0xA9DF,"Javanese"),
(0xA9E0,0xA9FF,"Myanmar Extended-B"),
(0xAA00,0xAA5F,"Cham"),
(0xAA60,0xAA7F,"Myanmar Extended-A"),
(0xAA80,0xAADF,"Tai Viet"),
(0xAAE0,0xAAFF,"Meetei Mayek Extensions"),
(0xAB00,0xAB2F,"Ethiopic Extended-A"),
(0xAB30,0xAB6F,"Latin Extended-E"),
(0xAB70,0xABBF,"Cherokee Supplement"),
(0xABC0,0xABFF,"Meetei Mayek"),
(0xAC00,0xD7AF,"Hangul Syllables"),
(0xD7B0,0xD7FF,"Hangul Jamo Extended-B"),
(0xD800,0xDB7F,"High Surrogates"),
(0xDB80,0xDBFF,"High Private Use Surrogates"),
(0xDC00,0xDFFF,"Low Surrogates"),
(0xE000,0xF8FF,"Private Use Area"),
(0xF900,0xFAFF,"CJK Compatibility Ideographs"),
(0xFB00,0xFB4F,"Alphabetic Presentation Forms"),
(0xFB50,0xFDFF,"Arabic Presentation Forms-A"),
(0xFE00,0xFE0F,"Variation Selectors"),
(0xFE10,0xFE1F,"Vertical Forms"),
(0xFE20,0xFE2F,"Combining Half Marks"),
(0xFE30,0xFE4F,"CJK Compatibility Forms"),
(0xFE50,0xFE6F,"Small Form Variants"),
(0xFE70,0xFEFF,"Arabic Presentation Forms-B"),
(0xFF00,0xFFEF,"Halfwidth and Fullwidth Forms"),
(0xFFF0,0xFFFF,"Specials"),
(0x10000,0x1007F,"Linear B Syllabary"),
(0x10080,0x100FF,"Linear B Ideograms"),
(0x10100,0x1013F,"Aegean Numbers"),
(0x10140,0x1018F,"Ancient Greek Numbers"),
(0x10190,0x101CF,"Ancient Symbols"),
(0x101D0,0x101FF,"Phaistos Disc"),
(0x10280,0x1029F,"Lycian"),
(0x102A0,0x102DF,"Carian"),
(0x102E0,0x102FF,"Coptic Epact Numbers"),
(0x10300,0x1032F,"Old Italic"),
(0x10330,0x1034F,"Gothic"),
(0x10350,0x1037F,"Old Permic"),
(0x10380,0x1039F,"Ugaritic"),
(0x103A0,0x103DF,"Old Persian"),
(0x10400,0x1044F,"Deseret"),
(0x10450,0x1047F,"Shavian"),
(0x10480,0x104AF,"Osmanya"),
(0x104B0,0x104FF,"Osage"),
(0x10500,0x1052F,"Elbasan"),
(0x10530,0x1056F,"Caucasian Albanian"),
(0x10600,0x1077F,"Linear A"),
(0x10800,0x1083F,"Cypriot Syllabary"),
(0x10840,0x1085F,"Imperial Aramaic"),
(0x10860,0x1087F,"Palmyrene"),
(0x10880,0x108AF,"Nabataean"),
(0x108E0,0x108FF,"Hatran"),
(0x10900,0x1091F,"Phoenician"),
(0x10920,0x1093F,"Lydian"),
(0x10980,0x1099F,"Meroitic Hieroglyphs"),
(0x109A0,0x109FF,"Meroitic Cursive"),
(0x10A00,0x10A5F,"Kharoshthi"),
(0x10A60,0x10A7F,"Old South Arabian"),
(0x10A80,0x10A9F,"Old North Arabian"),
(0x10AC0,0x10AFF,"Manichaean"),
(0x10B00,0x10B3F,"Avestan"),
(0x10B40,0x10B5F,"Inscriptional Parthian"),
(0x10B60,0x10B7F,"Inscriptional Pahlavi"),
(0x10B80,0x10BAF,"Psalter Pahlavi"),
(0x10C00,0x10C4F,"Old Turkic"),
(0x10C80,0x10CFF,"Old Hungarian"),
(0x10D00,0x10D3F,"Hanifi Rohingya"),
(0x10E60,0x10E7F,"Rumi Numeral Symbols"),
(0x10E80,0x10EBF,"Yezidi"),
(0x10F00,0x10F2F,"Old Sogdian"),
(0x10F30,0x10F6F,"Sogdian"),
(0x10FB0,0x10FDF,"Chorasmian"),
(0x10FE0,0x10FFF,"Elymaic"),
(0x11000,0x1107F,"Brahmi"),
(0x11080,0x110CF,"Kaithi"),
(0x110D0,0x110FF,"Sora Sompeng"),
(0x11100,0x1114F,"Chakma"),
(0x11150,0x1117F,"Mahajani"),
(0x11180,0x111DF,"Sharada"),
(0x111E0,0x111FF,"Sinhala Archaic Numbers"),
(0x11200,0x1124F,"Khojki"),
(0x11280,0x112AF,"Multani"),
(0x112B0,0x112FF,"Khudawadi"),
(0x11300,0x1137F,"Grantha"),
(0x11400,0x1147F,"Newa"),
(0x11480,0x114DF,"Tirhuta"),
(0x11580,0x115FF,"Siddham"),
(0x11600,0x1165F,"Modi"),
(0x11660,0x1167F,"Mongolian Supplement"),
(0x11680,0x116CF,"Takri"),
(0x11700,0x1173F,"Ahom"),
(0x11800,0x1184F,"Dogra"),
(0x118A0,0x118FF,"Warang Citi"),
(0x11900,0x1195F,"Dives Akuru"),
(0x119A0,0x119FF,"Nandinagari"),
(0x11A00,0x11A4F,"Zanabazar Square"),
(0x11A50,0x11AAF,"Soyombo"),
(0x11AC0,0x11AFF,"Pau Cin Hau"),
(0x11C00,0x11C6F,"Bhaiksuki"),
(0x11C70,0x11CBF,"Marchen"),
(0x11D00,0x11D5F,"Masaram Gondi"),
(0x11D60,0x11DAF,"Gunjala Gondi"),
(0x11EE0,0x11EFF,"Makasar"),
(0x11FB0,0x11FBF,"Lisu Supplement"),
(0x11FC0,0x11FFF,"Tamil Supplement"),
(0x12000,0x123FF,"Cuneiform"),
(0x12400,0x1247F,"Cuneiform Numbers and Punctuation"),
(0x12480,0x1254F,"Early Dynastic Cuneiform"),
(0x13000,0x1342F,"Egyptian Hieroglyphs"),
(0x13430,0x1343F,"Egyptian Hieroglyph Format Controls"),
(0x14400,0x1467F,"Anatolian Hieroglyphs"),
(0x16800,0x16A3F,"Bamum Supplement"),
(0x16A40,0x16A6F,"Mro"),
(0x16AD0,0x16AFF,"Bassa Vah"),
(0x16B00,0x16B8F,"Pahawh Hmong"),
(0x16E40,0x16E9F,"Medefaidrin"),
(0x16F00,0x16F9F,"Miao"),
(0x16FE0,0x16FFF,"Ideographic Symbols and Punctuation"),
(0x17000,0x187FF,"Tangut"),
(0x18800,0x18AFF,"Tangut Components"),
(0x18B00,0x18CFF,"Khitan Small Script"),
(0x18D00,0x18D8F,"Tangut Supplement"),
(0x1B000,0x1B0FF,"Kana Supplement"),
(0x1B100,0x1B12F,"Kana Extended-A"),
(0x1B130,0x1B16F,"Small Kana Extension"),
(0x1B170,0x1B2FF,"Nushu"),
(0x1BC00,0x1BC9F,"Duployan"),
(0x1BCA0,0x1BCAF,"Shorthand Format Controls"),
(0x1D000,0x1D0FF,"Byzantine Musical Symbols"),
(0x1D100,0x1D1FF,"Musical Symbols"),
(0x1D200,0x1D24F,"Ancient Greek Musical Notation"),
(0x1D2E0,0x1D2FF,"Mayan Numerals"),
(0x1D300,0x1D35F,"Tai Xuan Jing Symbols"),
(0x1D360,0x1D37F,"Counting Rod Numerals"),
(0x1D400,0x1D7FF,"Mathematical Alphanumeric Symbols"),
(0x1D800,0x1DAAF,"Sutton SignWriting"),
(0x1E000,0x1E02F,"Glagolitic Supplement"),
(0x1E100,0x1E14F,"Nyiakeng Puachue Hmong"),
(0x1E2C0,0x1E2FF,"Wancho"),
(0x1E800,0x1E8DF,"Mende Kikakui"),
(0x1E900,0x1E95F,"Adlam"),
(0x1EC70,0x1ECBF,"Indic Siyaq Numbers"),
(0x1ED00,0x1ED4F,"Ottoman Siyaq Numbers"),
(0x1EE00,0x1EEFF,"Arabic Mathematical Alphabetic Symbols"),
(0x1F000,0x1F02F,"Mahjong Tiles"),
(0x1F030,0x1F09F,"Domino Tiles"),
(0x1F0A0,0x1F0FF,"Playing Cards"),
(0x1F100,0x1F1FF,"Enclosed Alphanumeric Supplement"),
(0x1F200,0x1F2FF,"Enclosed Ideographic Supplement"),
(0x1F300,0x1F5FF,"Miscellaneous Symbols and Pictographs"),
(0x1F600,0x1F64F,"Emoticons (Emoji)"),
(0x1F650,0x1F67F,"Ornamental Dingbats"),
(0x1F680,0x1F6FF,"Transport and Map Symbols"),
(0x1F700,0x1F77F,"Alchemical Symbols"),
(0x1F780,0x1F7FF,"Geometric Shapes Extended"),
(0x1F800,0x1F8FF,"Supplemental Arrows-C"),
(0x1F900,0x1F9FF,"Supplemental Symbols and Pictographs"),
(0x1FA00,0x1FA6F,"Chess Symbols"),
(0x1FA70,0x1FAFF,"Symbols and Pictographs Extended-A"),
(0x1FB00,0x1FBFF,"Symbols for Legacy Computing"),
(0x20000,0x2A6DF,"CJK Unified Ideographs Extension B"),
(0x2A700,0x2B73F,"CJK Unified Ideographs Extension C"),
(0x2B740,0x2B81F,"CJK Unified Ideographs Extension D"),
(0x2B820,0x2CEAF,"CJK Unified Ideographs Extension E"),
(0x2CEB0,0x2EBEF,"CJK Unified Ideographs Extension F"),
(0x2F800,0x2FA1F,"CJK Compatibility Ideographs Supplement"),
(0x30000,0x3134F,"CJK Unified Ideographs Extension G"),
(0xE0000,0xE007F,"Tags"),
(0xE0100,0xE01EF,"Variation Selectors Supplement"),
(0xF0000,0xFFFFF,"Supplementary Private Use Area-A"),
(0x100000,0x10FFFF,"Supplementary Private Use Area-B"),
]


class UnicodepointIndex(Enum):

    ENCLOSED_ALPHANUMERICS = 82
    SMALL_FORM_VARIANTS = 159
    HALFWIDTH_AND_FULLWIDTH_FORMS = 161


_, breakpoints, names = zip(*unicode_mapping)

@cache(maxsize=None)
def index(c: Char) -> int:
    code = ord(c)
    i = bisect(breakpoints, code)
    return i-1 if code == breakpoints[i-1] else i

def is_enclosed_alphanumerics(c: Char):
    return UnicodepointIndex.ENCLOSED_ALPHANUMERICS.value == index(c)

def is_small_form_variants(c: Char):
    return UnicodepointIndex.SMALL_FORM_VARIANTS.value == index(c)

def is_halfwidth_and_fullwidth_forms(c: Char):
    return UnicodepointIndex.HALFWIDTH_AND_FULLWIDTH_FORMS.value == index(c)



# encoding: utf-8

import os
import numpy as np
import pandas as pd

from typing import List, Union, Tuple, Optional
from collections import OrderedDict
from collections.abc import Iterable

from ..util import logging
logger = logging.getLogger(__name__)

# Define type aliases
Char = str
CharsInput = List[Char]
CharsOutput = CharsInput

Text = str
TextInput = List[Text]

Token = str
TokenInput = List[Token]

Vocab = OrderedDict

# Init default value
IS_CHAR_DEFAULT = True
VOCAB_FILES_NAMES = {"vocab_file": "vocab.txt"}


class VocabularyBase:

    def __init__(self, vocab_file=None, **kwargs):
        self.vocab = OrderedDict() if vocab_file is None else self.load_vocab(vocab_file)
        self.special_tokens = []
        self.unk_token = '[UNK]'
        self.add_tokens(self.unk_token, special_tokens=True)

    def __contains__(self, token: Token) -> bool:
        return token in self.vocab

    def __eq__(self, other):
        if isinstance(other, OrderedDict):
            return self.vocab == other
        return self == other

    def items(self):
        return self.vocab.items()

    def is_special(self, token: Token) -> bool:
        return token in self.special_tokens

    @property
    def vocab_size(self):
        return len(self.vocab)

    def add_tokens(self, tokens: Union[Token, TokenInput], special_tokens: bool = False):

        def _add_tokens(token: Token, special_tokens: bool = False):
            if special_tokens and token not in self.special_tokens:
                self.special_tokens.append(token)
            if token not in self.vocab:
                self.vocab[token] = self.vocab_size

        if isinstance(tokens, str):
            _add_tokens(tokens, special_tokens=special_tokens)
            logger.debug(f'vocab size: {self.vocab_size}')
        elif isinstance(tokens, (list, tuple, Iterable, np.ndarray, pd.Series)):
            for token in tokens:
                _add_tokens(token, special_tokens=special_tokens)
            logger.debug(f'vocab size: {self.vocab_size}')
        else:
            raise ValueError(
                f"type of tokens unknown: {type(tokens)}. "
                f"Should be one of a str, list, tuple, set, Iterable, np.ndarray or pd.Series."
            )   

    def create_vocab(self, texts_or_path: Union[List[CharsInput], TextInput, os.PathLike], **kwargs):
        # TODO: support from counter

        is_char = kwargs.get('is_char', IS_CHAR_DEFAULT)

        def _create_vocab_with_special(texts: Union[List[CharsInput], TextInput], **kwargs):
            if is_char:
                self._create_vocab(texts, **kwargs)
            else:
                new_texts = []
                for text in texts:
                    if isinstance(text, str):
                        new_texts.append(text)
                    elif isinstance(text, (list, tuple, np.ndarray, pd.Series)):
                        for t in text:
                            if not self.is_special(t):
                                new_texts.append(t)
                    else:
                        raise ValueError(
                            f"type of text unknown: {type(text)}. "
                            f"Should be one of a str, list, tuple, np.ndarray, pd.Series."
                        )   
                self._create_vocab(new_texts, **kwargs)

        def _create_vocab_byfile(p, **kwargs):
            with p.open() as f:
                texts = f.readlines()
            _create_vocab_with_special(texts, **kwargs)

        if isinstance(texts_or_path, (list, tuple, np.ndarray, pd.Series)):
            _create_vocab_with_special(texts_or_path, **kwargs)
        else:
            p = Path(texts_or_path)
            if p.is_file():
                self._create_vocab_byfile(p, **kwargs)
            elif p.is_dir():
                # TODO: in 1 file
                for _ in p.iterdir(): self._create_vocab_byfile(_, **kwargs)
            else:
                raise ValueError(
                    f"type of texts_or_path unknown: {type(texts_or_path)}. "
                    f"Should be one of a str, list, tuple, np.ndarray, pd.Series, os.PathLike."
                )   

    def _create_vocab(self, texts: Union[List[CharsInput], TextInput], **kwargs):
        raise NotImplementedError

    def save_vocabulary(self, save_directory: str, filename_prefix: Optional[str] = None) -> Tuple[str]:
        index = 0
        os.makedirs(save_directory, exist_ok=True)
        vocab_file = os.path.join(
            save_directory, (filename_prefix + "-" if filename_prefix else "") + VOCAB_FILES_NAMES["vocab_file"]
        )
        with open(vocab_file, "w", encoding="utf-8") as writer:
            for token, token_index in self.vocab.items():
                if index != token_index:
                    logger.warning(
                        "Saving vocabulary to {}: vocabulary indices are not consecutive."
                        " Please check that the vocabulary is not corrupted!".format(vocab_file)
                    )
                    index = token_index
                writer.write(token + "\n")
                index += 1
        return (vocab_file,)

    def load_vocab(self, save_directory: str, filename_prefix: Optional[str] = None) -> Vocab:
        """Loads a vocabulary file into a dictionary."""
        vocab_file = os.path.join(
            save_directory, (filename_prefix + "-" if filename_prefix else "") + VOCAB_FILES_NAMES["vocab_file"]
        )
        with open(vocab_file, "r", encoding="utf-8") as reader:
            tokens = reader.readlines()
        for index, token in enumerate(tokens):
            token = token.rstrip("\n")
            self.vocab[token] = index



# encoding: utf-8
#
# ref: https://unicode-table.com/en

from typing import List, Union, Tuple, Optional
from .unicode_mapping import is_enclosed_alphanumerics
from .unicode_mapping import is_small_form_variants
from .unicode_mapping import is_halfwidth_and_fullwidth_forms

# Define type aliases
Char = str


class NormalizerBase:

    def enclosed_alphanumerics_to_basic_latin(self, c: Char) -> Char:

        def numeric(code: int) -> int:
            code = zero(code)
            if isinstance(code,int) and code == code: code = one_to_nine(code)
            if isinstance(code,int) and code == code: code = ten_to_nineteen(code)
            if isinstance(code,int) and code == code: code = twenty(code)
            return code

        def zero(code: int) -> int:
            return 0x0030 if 0x24ea == code \
                or 0x24ff == code \
                else code

        def one_to_nine(code: int) -> int:
            if 0x2460 <= code and 0x2468 >= code:
                return code-0x242f 
            elif 0x2474 <= code and 0x247c >= code:
                return code-0x2443
            elif 0x2488 <= code and 0x2490 >= code:
                return code-0x2457
            elif 0x24f5 <= code and 0x24fd >= code:
                return code-0x24c4
            return code

        def ten_to_nineteen(code: int) -> int:
            if 0x2469 <= code and 0x2472 >= code:
                return [0x0031, code-0x2439]
            elif 0x247d <= code and 0x2486 >= code:
                return [0x0031, code-0x244d]
            elif 0x2491 <= code and 0x249a >= code:
                return [0x0031, code-0x2461]
            elif 0x24eb <= code and 0x24f3 >= code: # 11-19
                return [0x0031, code-0x24ba]
            elif 0x24fe == code: # 10
                return [0x0031, 0x0030]
            return code

        def twenty(code: int) -> int:
            return [0x0032, 0x0030] if 0x2473 == code \
                or 0x2487 == code \
                or 0x249b == code \
                or 0x24f4 == code \
                else code

        def alphabet(code: int) -> int:
            code = alphabet_lower(code)
            if isinstance(code,int) and code == code: code = alphabet_upper(code)
            return code

        def alphabet_lower(code: int) -> int:
            if 0x249c <= code and 0x24b5 >= code:
                return code-0x243b
            elif 0x24d0 <= code and 0x24e9 >= code:
                return code-0x246f
            return code

        def alphabet_upper(code: int) -> int:
            if 0x24b6 <= code and 0x24cf >= code:
                return code-0x2475
            return code

        code = ord(c)
        code = numeric(code)
        if isinstance(code,int) and code == code: code = alphabet(code)
        if isinstance(code, int):
            return chr(code)
        elif isinstance(code, list):
            return ''.join([chr(c) for c in code])
        else:
            raise ValueError(
                f"type of code unknown: {type(code)}. "
                f"Should be one of a str or list."
            )  

    def small_form_variants_to_basic_latin(self, c: Char) -> Char:
        mapping = {
            0xfe50 : 0x002c,
            0xfe51 : 0x3001, # CJK Symbols and Punctuation
            0xfe52 : 0x002e,

            0xfe54 : 0x003b,
            0xfe55 : 0x003a,
            0xfe56 : 0x003f,
            0xfe57 : 0x0021,
            0xfe58 : 0x2014, # General Punctuation
            0xfe59 : 0x0028,
            0xfe5a : 0x0029,
            0xfe5b : 0x007b,
            0xfe5c : 0x007d,
            0xfe5d : 0x005b,
            0xfe5e : 0x005d,
            0xfe5f : 0x0023,
            0xfe60 : 0x0026,
            0xfe61 : 0x002a,
            0xfe62 : 0x002b,
            0xfe63 : 0x002d,
            0xfe64 : 0x003c,
            0xfe65 : 0x003e,
            0xfe66 : 0x003d,

            0xfe68 : 0x005c,
            0xfe69 : 0x0024,
            0xfe6a : 0x0025,
            0xfe6b : 0x0040,
        }
        code = mapping.get(ord(c))
        return chr(code) if code else c

    def halfwidth_and_fullwidth_forms_to_basic_latin(self, c: Char) -> Char:
        code = ord(c)
        if 0xff01 <= code and 0xff5e >= code:
            return chr(code-0xfee0)
        return c

    def __call__(self, c: Char) -> Char:
        if is_enclosed_alphanumerics(c):
            return self.enclosed_alphanumerics_to_basic_latin(c)
        elif is_small_form_variants(c):
            return self.small_form_variants_to_basic_latin(c)
        elif is_halfwidth_and_fullwidth_forms(c):
            return self.halfwidth_and_fullwidth_forms_to_basic_latin(c)
        return c


class ChineseNormalizerBase(NormalizerBase):

    def cjk_compatibility_to_normal(self, c: Char) -> Char:
        # TODO:
        pass

    def __call__(self, c: Char) -> Char:
        # TODO:
        pass



# encoding: utf-8

import os
import copy
import numpy as np
import pandas as pd

from typing import List, Union, Tuple, Optional
from collections import OrderedDict
from collections.abc import Iterable


# Define type aliases
Text = str

Token = str
TokenInput = List[str]

# Slow tokenizers used to be saved in three separated files
TOKENIZER_CONFIG_FILE = "tokenizer_config.json"


class TokenizerBase:
    
    def __init__(self, **kwargs):
        # inputs and kwargs for saving and re-loading (see ``from_pretrained`` and ``save_pretrained``)
        self.init_inputs = ()
        self.init_kwargs = copy.deepcopy(kwargs)

    def tokenize(self, text: Text, **kwargs) -> List[str]:
        raise NotImplementedError

    def save_pretrained(
        self, 
        save_directory: Union[str, os.PathLike],
        filename_prefix: Optional[str] = None
    ) -> Tuple[str]:

        if os.path.isfile(save_directory):
            logger.error("Provided path ({}) should be a directory, not a file".format(save_directory))
            return
        os.makedirs(save_directory, exist_ok=True)

        tokenizer_config_file = os.path.join(
            save_directory, (filename_prefix + "-" if filename_prefix else "") + TOKENIZER_CONFIG_FILE
        )

        tokenizer_config = copy.deepcopy(self.init_kwargs)
        if len(self.init_inputs) > 0:
            tokenizer_config["init_inputs"] = copy.deepcopy(self.init_inputs)
        for file_id in self.vocab_files_names.keys():
            tokenizer_config.pop(file_id, None)

        with open(tokenizer_config_file, "w", encoding="utf-8") as f:
            f.write(json.dumps(tokenizer_config, ensure_ascii=False))
        # logger.info(f"tokenizer config file saved in {tokenizer_config_file}")
        # TODO:
        pass

    @classmethod
    def from_pretrained(cls, pretrained_model_name_or_path: Union[str, os.PathLike], *init_inputs, **kwargs):
        # Prepare tokenizer initialization kwargs
        # TODO:
        filename_prefix = None
        tokenizer_config_file = os.path.join(
            save_directory, (filename_prefix + "-" if filename_prefix else "") + TOKENIZER_CONFIG_FILE
        )
        with open(tokenizer_config_file, encoding="utf-8") as tokenizer_config_handle:
            init_kwargs = json.load(tokenizer_config_handle)
        saved_init_inputs = init_kwargs.pop("init_inputs", ())
        if not init_inputs:
            init_inputs = saved_init_inputs

        # Update with newly provided kwargs
        init_kwargs.update(kwargs)

        # Instantiate tokenizer.
        try:
            tokenizer = cls(*init_inputs, **init_kwargs)
        except OSError:
            raise OSError(
                "Unable to load vocabulary from file. "
                "Please check that the provided vocabulary is accessible and not corrupted."
            )

        return tokenizer



# encoding: utf-8

from .normalizer import NormalizerBase
from .vocabulary import VocabularyBase
from .tokenizer import TokenizerBase

from .pipeline import (
    Step,
    Pipeline
)

from .algorithm import (
    BPE,
    VocabularyBPE,
    TokenizerBPE,
)



import logging

# set up logging to file - see previous section for more details
logging.basicConfig(level=logging.DEBUG,
                    format='%(asctime)s %(name)-12s %(levelname)-8s %(message)s',
                    datefmt='%m-%d %H:%M')

# define a Handler which writes INFO messages or higher to the sys.stderr
console = logging.StreamHandler()
console.setLevel(logging.INFO)

# set a format which is simpler for console use
formatter = logging.Formatter('%(name)-12s: %(levelname)-8s %(message)s')

# tell the handler to use this format
console.setFormatter(formatter)

# add the handler to the root logger
logging.getLogger('').addHandler(console)

# Now, we can log to the root logger, or any other logger. First the root...
# logging.info('Jackdaws love my big sphinx of quartz.')

# Now, define a couple of other loggers which might represent areas in your
# application:

# logger = logging.getLogger('myapp.area1')



# encoding: utf-8

import numpy as np
import pandas as pd

from enum import Enum
from itertools import chain
from pathlib import Path
from collections import Counter
from typing import List, Dict, Union, Tuple, Optional

from ...vocabulary import VocabularyBase
from ...tokenizer import TokenizerBase


# Define type aliases
Vocabulary = VocabularyBase


class MethodSupport(Enum):

    MOST_FREQ_1 = 'most_freq_1'
    MOST_FREQ_ALL = 'most_freq_all'

    LESS_FREQ_1 = 'less_freq_1'
    LESS_FREQ_ALL = 'less_freq_all'

    MOST_LESS_FREQ_1 = 'most_less_freq_1'
    MOST_LESS_FREQ_ALL = 'most_less_freq_all'

    TFIDF_1 = 'tfidf_1'
    TFIDF_ALL = 'tfidf_all'


METHOD_MOST_FREQ = [
    MethodSupport.MOST_FREQ_1,
    MethodSupport.MOST_FREQ_ALL,
]

METHOD_LESS_FREQ = [
    MethodSupport.MOST_FREQ_1,
    MethodSupport.MOST_FREQ_ALL,
]

METHOD_MOST_LESS_FREQ = [
    MethodSupport.MOST_LESS_FREQ_1,
    MethodSupport.MOST_LESS_FREQ_ALL,
]

METHOD_FREQ = METHOD_MOST_FREQ + METHOD_LESS_FREQ + METHOD_MOST_LESS_FREQ

METHOD_TFIDF = [
    MethodSupport.TFIDF_1,
    MethodSupport.TFIDF_ALL,
]


class Priority(Enum):
    """
    Not used on METHOD ALL
    """

    MAX_LENGTH_FIRST = 'max_length_first'
    MIN_LENGTH_FIRST = 'min_lenght_first'


# Define type aliases
Char = str
CharsInput = List[Char]
CharsOutput = CharsInput

Token = Char
TokenId = int
TokenInput = List[Token]
TokenOutput = TokenInput

Text = str
TextInput = List[Text]


# Init default value
MAX_LENGTH_DEFAULT = 7
METHOD_DEFAULT = MethodSupport.MOST_FREQ_1
PRIORITY_DEFAULT = Priority.MAX_LENGTH_FIRST

VOCAB_SIZE_DEFAULT = 1e4
KEEP_SINGLE_DEFAULT = True
CONVERT_UNKNOWN_DEFAULT = True


class BPE:

    def to_char(self, texts: Union[Text, TextInput, List[CharsInput]]) -> Union[CharsInput, List[CharsInput]]:
        # TODO: about space and control
        # TODO: implement by normalizer
        return texts

    def to_pairs(self, chars: Union[Text, CharsInput, List[CharsInput]], **kwargs) -> Union[CharsOutput, List[CharsOutput]]:
        """
        Get all possible pairs, limit by {max_length}
        """
        if isinstance(chars, str):
            return self._to_pairs(chars)
        elif isinstance(chars, (list, tuple, np.ndarray, pd.Series)):
            first_element = chars[0]
            if isinstance(first_element, (str, list, tuple, np.ndarray, pd.Series)):
                pairs = [self._to_pairs(_) for _ in chars]
                return [_ for _ in pairs if _]
            else:
                raise ValueError(
                    f"type of chars element unknown: {type(first_element)}. "
                    f"Should be one of a str, list, tuple, np.ndarray or pd.Series."
                )
        else:
            raise ValueError(
                f"type of chars unknown: {type(chars)}. "
                f"Should be one of a list, tuple, np.ndarray or pd.Series."
            )

    def _to_pairs(self, chars: Union[Text, CharsInput], **kwargs) -> CharsOutput:
        max_length = kwargs.get('max_length', MAX_LENGTH_DEFAULT)
        if isinstance(chars, str) and 2 >= len(chars):
            return [chars]
        elif 2 > len(chars):
            return []
        else:
            pairs = [chars[i]+chars[i+1] for i in range(len(chars)-1)]
            return [p for p in pairs if max_length >= len(p)]

    def get_pairs(self, chars: Union[CharsInput, List[CharsInput]], **kwargs) -> Union[CharsOutput, List[CharsOutput]]:
        method = kwargs.get('method', METHOD_DEFAULT)

        if method in METHOD_FREQ:
            return self._get_pairs_freq(chars, **kwargs)
        elif method in METHOD_TFIDF:
            return self._get_pairs_tfidf(chars, **kwargs)
        else:
            raise ValueError(
                f"type of method unknown: {method}. "
                f"Should be one of a {METHOD_FREQ} or {METHOD_TFIDF}."
            )

    def _get_pairs_freq(self, chars: Union[CharsInput, List[CharsInput]], **kwargs) -> Union[CharsOutput, List[CharsOutput]]:
        raise NotImplementedError

    def _get_pairs_tfidf(self, chars: Union[CharsInput, List[CharsInput]], **kwargs) -> Union[CharsOutput, List[CharsOutput]]:
        raise NotImplementedError

    def update(
        self,
        chars: Union[Text, CharsInput, List[CharsInput]],
        pairs: Union[Char, CharsOutput]
    ) -> Union[CharsOutput, List[CharsOutput]]:
        if isinstance(chars, str):
            return self._update(chars, pairs)
        elif isinstance(chars, (list, tuple, np.ndarray, pd.Series)):
            first_element = chars[0]
            if isinstance(first_element, (str, list, tuple, np.ndarray, pd.Series)):
                return [self._update(_, pairs) for _ in chars]
            else:
                raise ValueError(
                    f"type of chars element unknown: {type(first_element)}. "
                    f"Should be one of a str, list, tuple, np.ndarray or pd.Series."
                )
        else:
            raise ValueError(
                f"type of chars unknown: {type(chars)}. "
                f"Should be one of a list, tuple, np.ndarray or pd.Series."
            )

    def _update(
        self,
        chars: Union[Text, CharsInput],
        pairs: Union[Char, CharsOutput]
    ) -> CharsOutput:

        if 1 >= len(chars):
            return chars
        else:
            if isinstance(chars, str): chars = list(chars)
            if isinstance(pairs, str): pairs = [pairs]
            for pair in pairs:
                new_chars = []
                i = 0
                while(i < len(chars)):
                    if pair == ''.join(chars[i:i+2]):
                        new_chars.append(pair)
                        i += 2
                        continue
                    new_chars.append(chars[i])
                    if i == (len(chars)-2):
                        new_chars.append(chars[i+1])
                        break
                    i += 1
                chars = new_chars
            return chars


class VocabularyBPE(VocabularyBase, BPE):

    def _get_pairs_freq(self, chars: List[CharsInput], **kwargs) -> List[CharsOutput]:
        """
        Get appropriate pair(s) by {method} and {priority}
        """
        method = kwargs.get('method', METHOD_DEFAULT)
        if method in METHOD_MOST_FREQ:
            return self._get_pairs_freq_most(chars, **kwargs)
        elif method in METHOD_LESS_FREQ:
            return self._get_pairs_freq_less(chars, **kwargs)
        elif method in METHOD_MOST_LESS_FREQ:
            return self._get_pairs_freq_most_less(chars, **kwargs)
        else:
            raise ValueError(
                f"type of method unknown: {method}. "
                f"Should be one of a {METHOD_FREQ}."
            )

    def _get_pairs_freq_most(self, chars: List[CharsInput], **kwargs) -> List[CharsOutput]:
        method = kwargs.get('method', METHOD_DEFAULT)
        if MethodSupport.MOST_FREQ_1 == method:
            return self._get_pairs_freq_most_1(chars, **kwargs)
        elif MethodSupport.MOST_FREQ_ALL == method:
            return self._get_pairs_freq_most_all(chars, **kwargs)
        else:
            raise ValueError(
                f"type of method unknown: {method}. "
                f"Should be one of a {MethodSupport.MOST_FREQ}."
            )

    def _get_pairs_freq_most_1(self, chars: List[CharsInput], **kwargs) -> List[CharsOutput]:
        return self._get_pairs_freq_most_all(chars)[0]

    def _get_pairs_freq_most_all(self, chars: List[CharsInput], **kwargs) -> List[CharsOutput]:
        counter = Counter(chain.from_iterable(chars))

        most_freq = counter.most_common(1)[0][1]
        pairs = []
        for (k, cnt) in  counter.most_common():
            if most_freq == cnt:
                pairs.append(k)
                continue
            break

        return self._order_pairs(pairs)

    def _get_pairs_freq_less(self, chars: List[CharsInput], **kwargs) -> List[CharsOutput]:
        # TODO:
        raise NotImplementedError

    def _get_pairs_freq_most_less(self, chars: List[CharsInput], **kwargs) -> List[CharsOutput]:
        # TODO:
        raise NotImplementedError

    def _get_pairs_tfidf(self, chars: List[CharsInput], **kwargs) -> List[CharsOutput]:
        """
        Get appropriate pair(s) by {method} and {priority}
        """
        # TODO:
        raise NotImplementedError

    def _order_pairs(self, pairs: CharsInput, **kwargs) -> CharsOutput:
        priority = kwargs.get('priority', PRIORITY_DEFAULT)
        
        if Priority.MAX_LENGTH_FIRST == priority:
            return sorted(pairs, key=len, reverse=True)
        elif Priority.MIN_LENGTH_FIRST == priority:
            return sorted(pairs, key=len)
        else:
            raise ValueError(
                f"type of priority unknown: {priority}. "
                f"Should be one of a {Priority.MAX_LENGTH_FIRST} or {Priority.MIN_LENGTH_FIRST}."
            )

    def _create_vocab(self, texts: Union[List[CharsInput], TextInput], **kwargs):
        vocab_size = kwargs.get('vocab_size', VOCAB_SIZE_DEFAULT)
        keep_single = kwargs.get('keep_single', KEEP_SINGLE_DEFAULT)

        chars = self.to_char(texts)

        if keep_single:
            self.add_tokens(sorted(list(set(chain.from_iterable(chars)))))

        while(self.vocab_size < vocab_size):
            pairs = self.to_pairs(chars)
            if not pairs: break
            pairs = self.get_pairs(pairs, **kwargs)
            self.add_tokens(pairs)
            chars = self.update(chars, pairs)


class TokenizerBPE(TokenizerBase, BPE):

    def __init__(self, vocab: Vocabulary=None, **kwargs):
        self.vocab = vocab
        super().__init__(**kwargs)

    def _get_pairs(self, chars: CharsInput, **kwargs) -> CharsOutput:
        method = kwargs.get('method', METHOD_DEFAULT)

        if (isinstance(method,MethodSupport) and method.value.endswith('1')) or (isinstance(method,str) and method.endswith('1')):
            for k, _ in self.vocab.items():
                if k in chars:
                    return k
            return ''

        return [k for k, _ in self.vocab.items() if k in chars]

    def _get_pairs_freq(self, chars: CharsInput, **kwargs) -> CharsOutput:
        """
        Get appropriate pair(s) by {method} and {priority} with importance of vocab order.
        """
        return self._get_pairs(chars, **kwargs)

    def _get_pairs_tfidf(self, chars: CharsInput, **kwargs) -> CharsOutput:
        """
        Get appropriate pair(s) by {method} and {priority} with importance of vocab order.
        """
        return self._get_pairs(chars, **kwargs)

    def tokenize(self, text: Text, **kwargs) -> TokenOutput:
        """
        Split a string into char, then merge pair char by ordered vocab until no more find.
        Replacing unknown tokens with the :obj:`unk_token`.
        """
        convert_unknown = kwargs.get('convert_unknown', CONVERT_UNKNOWN_DEFAULT)

        def merge(chars: CharsInput, **kwargs) -> CharsOutput:
            """
            Merge pair char by ordered vocab.
            """
            need_merge = True
            while(need_merge):
                need_merge, chars = _merge(chars, **kwargs)
            return chars

        def _merge(chars: CharsInput, **kwargs) -> CharsOutput:
            """
            Use trained vocabulary makes it different from get_pairs().
            The importance of Token has already been recorded by vocab order.
            """
            pairs = [c for c in self._to_pairs(chars, **kwargs) if c in self.vocab]
            if not pairs: return False, chars

            pairs = self.get_pairs(pairs)
            chars = self.update(chars, pairs)
            return True, chars

        def to_unknown(chars: CharsInput, vocab: Vocabulary):
            # TODO: to be implement with transformers
            return [c if c in self.vocab else self.vocab.unk_token for c in chars]

        chars = self.to_char(text)
        chars = merge(chars)
        if convert_unknown:
            chars = to_unknown(chars, self.vocab)
        return chars



# encoding: utf-8

import uuid
import json
import numpy as np
import pandas as pd

from collections import OrderedDict
from typing import List, Set, Dict, Union, Tuple, Optional, Any, Callable


# Define type aliases
Name = str
Required = List[Name]
Function = Callable[[Any],Any]
Args = List
Kwargs = Dict
FunctionTuple = Union[
    Tuple[Function],
    Tuple[Function, Args],
    Tuple[Function, Kwargs],
    Tuple[Function, Args, Kwargs],
    Tuple[Function, Args, Required],
    Tuple[Function, Kwargs, Required],
    Tuple[Function, Args, Kwargs, Required],
]

# Init default value
DEPEND_PREV_DEFAULT = True


class Step:

    def __init__(self, *args, **kwargs):
        self._id = str(uuid.uuid4())
        self.name = kwargs.pop('_name', self._id) # TODO: for required specify
        self.required = kwargs.pop('_required', [])
        self.args = args
        self.kwargs = kwargs
        self.outputs = None

    @property
    def info(self):
        return {
            'id': self._id,
            'name': self.name,
            'required': self.required,
            'args': self.args,
            'kwargs': self.kwargs,
            'outputs': self.outputs
        }

    @classmethod
    def from_callable(cls, _callable: Union[Callable, FunctionTuple]):
        obj = cls()
        if isinstance(_callable, Callable):
            obj._execute = _callable
        elif isinstance(_callable, (list, tuple, np.ndarray, pd.Series)):
            if 1 == len(_callable):
                obj._execute = _callable[0]
            elif 2 == len(_callable):
                function, _ = _callable
                if isinstance(function, Callable):
                    obj._execute = function
                    if isinstance(_, (list, tuple)):
                        obj.args = _
                    elif isinstance(_, dict):
                        obj.kwargs = _
                    else:
                        raise ValueError(
                            f"type of args or kwargs unknown: {type(_)}. "
                            f"Should be list, tuple or dict."
                        ) 
                else:
                    raise ValueError(
                        f"type of function unknown: {type(function)}. "
                        f"Should be Callable."
                    ) 
            elif 3 == len(_callable):
                function, t1, t2 = _callable
                if isinstance(function, Callable):
                    obj._execute = function
                    if isinstance(t1, (list, tuple)):
                        if isinstance(t2, dict):
                            obj.args = t1
                            obj.kwargs = t2
                        elif isinstance(t2, (list, tuple)):
                            obj.args = t1
                            obj.required = t2
                        else:
                            raise ValueError(
                                f"type of args, kwargs, required unknown: {type(t1), type(t2)}. "
                                f"Should be (list, dict), (list, list)."
                            ) 
                    elif isinstance(t1, dict):
                        if isinstance(t2, (list, tuple)):
                            obj.kwargs = t1
                            obj.required = t2
                        else:
                            raise ValueError(
                                f"type of kwargs, required unknown: {type(t1), type(t2)}. "
                                f"Should be (dict, dict)."
                            ) 
                    else:
                        raise ValueError(
                            f"type of args or kwargs unknown: {type(t1)}. "
                            f"Should be list, tuple or dict."
                        ) 
                else:
                    raise ValueError(
                        f"type of function unknown: {type(function)}. "
                        f"Should be Callable."
                    ) 
            elif 4 == len(_callable):
                obj._execute, obj.args, obj.kwargs, obj.required = _callable
            else:
                raise ValueError(
                    f"length of callable wrong: {len(_callable)}. "
                    f"Should be 1 to 4 for Tuple[Function, Args, Kwargs, Required]"
                ) 
        else:
            raise ValueError(
                f"type of callable unknown: {type(_callable)}. "
                f"Should be one of a Callable, list, tuple, np.ndarray or pd.Series."
            ) 
        return obj

    def meet(self, ready: Set[Name]):
        return not (set(self.required)-set(ready)) 

    def __call__(self, *args, **kwargs):
        kwargs.update(self.kwargs)
        self.outputs = self._execute(*(list(args)+list(self.args)), **kwargs)
        return self.outputs

    def _execute(self, *args, **kwargs):
        """
        Args:
            args (:obj:`list`):
                The arguments from dependency steps.
            kwargs (:obj:`dict`):
                The arguments given by user directly.
        """
        raise NotImplementedError


class Pipeline:

    def __init__(self, callables: List[Union[Name, Step, FunctionTuple]], **kwargs):

        def build(callables: List[Union[Name, Step, FunctionTuple]], **kwargs) -> List[Step]:
            """
            Convert FunctionTuple to Steps.
            Parsing relations between Steps for not specified.
            """
            depend_prev = kwargs.get('depend_prev', DEPEND_PREV_DEFAULT)

            new_callables = []
            for i, step in enumerate(callables):
                if isinstance(step, str):
                    step = Step.from_callable(getattr(self, step))
                elif not isinstance(step,Step) and isinstance(step, Callable):
                    step = Step.from_callable(step)
                elif isinstance(step, tuple):
                    if isinstance(step[0], str):
                        step = list(step)
                        step[0] = getattr(self, step[0])
                        step = Step.from_callable(tuple(step))
                    elif not isinstance(step[0],Step) and isinstance(step[0], Callable):
                        step = Step.from_callable(step)
                    else:
                        raise ValueError(
                            f"type of step[0] unknown: {type(step[0])}. "
                            f"Should be str, Callable."
                        ) 
                elif isinstance(step, Step):
                    pass
                else:
                    raise ValueError(
                        f"type of step unknown: {type(step)}. "
                        f"Should be str, Callable, tuple or Step."
                    ) 
                if depend_prev and not step.required and 0 != i:
                    step.required = set([new_callables[-1]._id])
                new_callables.append(step)
            return new_callables

        self.wait = build(callables)
        self.start = True
        self.finish = OrderedDict()

    @property
    def info(self):
        return [step.info for step in (self.wait if self.wait else self.finish)]

    def __call__(self, *args, **kwargs):
        
        def required(step):
            return [self.finish[req].outputs for req in step.required]

        while(self.wait):
            step = self.wait.pop(0)
            if step.meet(self.finish_names):
                req = required(step)
                if self.start:
                    req = list(args) + req
                    self.start = False
                    step(*req, **kwargs)
                else:
                    step(*req)
                self.finish[step._id] = step
            else:
                self.wait.append(step)
        return self.finish.popitem()[1].outputs

    @property
    def finish_names(self):
        return set(self.finish.keys())
